/* See {float_image_paint.h}. */
/* Last edited on 2008-07-30 01:47:46 by stolfi */

#define _GNU_SOURCE
#include <math.h>
#include <limits.h>
#include <assert.h>
#include <string.h>
 
#include <bool.h>
#include <jsmath.h>
#include <affirm.h>
#include <float_image.h>
#include <float_image_paint.h>

void float_image_paint_sample(float_image_t *A, int c, int ix, int iy, float_image_func_t func, int m)
  {
    int nx = A->sz[1];
    int ny = A->sz[2];
    
    /* Clip the arguments to the valid range: */
    if ((ix < 0) || (ix >= nx)) { return; }
    if ((iy < 0) || (iy >= ny)) { return; }
    if (m < 0) { m = 0; }

    /* Get coordinates {xp,yp} of pixel center: */
    double xp = ix + 0.5;
    double yp = iy + 0.5;
        
    /* Sample the new image {func} over the pixel: */
    double sum_w = 0.0;   /* Sum of antialias weights of all subsamples. */
    double sum_wc = 0.0;  /* Sum of antialias weights of opaque subsamples. */
    double sum_wcv = 0.0; /* Sum of antialias weights times value of opaque subsamples. */
    int dx, dy;
    for (dx = -m; dx <= +m; dx++)
      { double xs = ((double)dx)/(m+1); /* X displacement of subpixel sample. */
        double xa = fabs(xs);
        double wx = (xa < 0.5 ? 1 - 2*xa*xa : 2*(1 - xa)*(1 - xa)); /* X weight factor. */
        for (dy = -m; dy <= m; dy++)
          { double ys = ((double)dy)/(m+1); /* Y displacement of subpixel sample. */
            double ya = fabs(ys);
            double wy = (ya < 0.5 ? 1 - 2*ya*ya : 2*(1 - ya)*(1 - ya)); /* Y weight factor. */
            double w = wx*wy; /* Subsample weight. */
            assert(w > 0.0);
            double v = func(xp + xs, yp + ys);  /* New image value. */
            if (! isnan(v))
              { sum_wc += w;
                sum_wcv += w*v;
              }
            sum_w += w;
          }
      }
    
    if (sum_wc > 0)
      { /* Mix the image sample: */
        assert(sum_w > 0.0);
        double w_old = 1.0 - sum_wc/sum_w;   
        float *sp = float_image_get_sample_address(A, c, ix, iy);
        double v_old = (*sp);
        (*sp) = w_old * v_old + sum_wcv/sum_w;
      }
  }

void float_image_paint_samples
  ( float_image_t *A, 
    int c, 
    int xLo, 
    int xHi,
    int yLo,
    int yHi, 
    float func (double x, double y),
    int m 
  )
  {
    int nx = A->sz[1];
    int ny = A->sz[2];
    /* Clip the arguments to the valid range: */
    if (xLo < 0) { xLo = 0; }
    if (xHi >= nx ) { xHi = nx-1; }
    if (yLo < 0) { yLo = 0; }
    if (yHi >= ny ) { yHi = ny-1; }
    if (m < 0) { m = 0; }
    /* Paint each pixel: */
    int ix, iy;
    for (ix = xLo; ix <= xHi; ix++)
      { for (iy = yLo; iy <=yHi; iy++)
          { float_image_paint_sample(A, c, ix, iy, func, m); }
      }
  }

void float_image_paint_dot
  ( float_image_t *A, 
    int c, 
    double xctr, 
    double yctr, 
    double rad,
    double hwd,
    bool_t round, 
    bool_t diagonal, 
    float vfill,
    float vdraw,
    int m
  )
  {
    int nx = A->sz[1];
    int ny = A->sz[2];
    
    /* If the drawing ink is invisible, ignore the line width: */
    if (isnan(vdraw)) { hwd = 0.0; }
    
    /* The signs of {rad} and {hwd} are irrelevant: */
    rad = fabs(rad);
    hwd = fabs(hwd);
    
    /* Determine the max extent {rMax} along any axis: */
    double rMax;
    if ((round) || (! diagonal))
      { rMax = rad + hwd; }
    else
      { rMax = rad*M_SQRT2 + hwd; }
    
    int xLo, xHi; float_image_get_index_range(xctr, rMax, nx, &xLo, &xHi);
    int yLo, yHi; float_image_get_index_range(yctr, rMax, ny, &yLo, &yHi);
    
    if (round)
      { /* Round dot. */
        double orad = rad + hwd;
        double irad = fmax(0.0, rad - hwd);
        double orad2 = orad*orad; /* Outer radius, squared. */
        double irad2 = irad*irad; /* Inner radius, squared. */ 
        
        auto float func_round(double x, double y);
    
        float func_round(double x, double y)
          { double dx = x - xctr;
            double dy = y - yctr;
            double dr2 = dx*dx + dy*dy;
            if (dr2 < irad2) 
              { return vfill; }
            else if (dr2 < orad2)
              { return vdraw; }
            else
              { return NAN; }
          }
        
        float_image_paint_samples(A, c, xLo, xHi, yLo, yHi, func_round, m);
      }
    else
      { /* Square dot. */
        double orad = rad + hwd;            /* Outer half-side. */
        double irad = fmax(0.0, rad - hwd); /* Inner half-side. */
        double hwd2 = hwd*hwd;              /* Corner radius, squared. */
        
        auto float func_square(double x, double y);
    
        float func_square(double x, double y)
          { double dx = x - xctr;
            double dy = y - yctr;
            if (diagonal)
              { /* Rotate by 45 degrees: */
                double tx = M_SQRT1_2 * (dx + dy);
                double ty = M_SQRT1_2 * (dy - dx); 
                dx = tx; dy = ty;
              }
            dx = fabs(dx);
            dy = fabs(dy);
            if ((dx < irad) && (dy < irad))
              { return vfill; }
            else if ((dx > orad) || (dy > orad))
              { return NAN; }
            else if ((dx < rad) || (dy < rad))
              { return vdraw; }
            else 
              { /* Round corner: */
                dx -= rad;
                dy -= rad;
                double dr2 = dx*dx + dy*dy;
                if (dr2 < hwd2) 
                  { return vdraw; }
                else
                  { return NAN; }
              }
          }
        
        float_image_paint_samples(A, c, xLo, xHi, yLo, yHi, func_square, m);
      }
  }
  
void float_image_get_index_range(double z, double r, int n, int *iLo, int *iHi)
  {
    /* Safety check for grossly out-of-domain coordinates: */
    z = fmin(n + r + 2, fmax(- r - 2, z)); 
    int lo = (int)floor(z - r - 0.4999999); if (lo < 0) { lo = 0; }
    int hi = (int)floor(z + r + 0.4999999); if (hi >= n) { hi = n - 1; }
    (*iLo) = lo; (*iHi) = hi;
  }

void float_image_paint_cross
  ( float_image_t *A, 
    int c, 
    double xctr, 
    double yctr, 
    double rad,
    double hwd, 
    bool_t diagonal, 
    float vdraw,
    int m
  )
  {
    int nx = A->sz[1];
    int ny = A->sz[2];
    
    /* If the drawing ink is invisible, there is nothing to do: */
    if (isnan(vdraw)) { return; }
    
    /* The signs of {rad} and {hwd} are irrelevant: */
    rad = fabs(rad);
    hwd = fabs(hwd);
    
    /* Determine the max extent {rMax} along any axis: */
    double rMax;
    if (! diagonal)
      { rMax = rad + hwd; }
    else
      { rMax = rad*M_SQRT1_2 + hwd; }
    
    int xLo, xHi; float_image_get_index_range(xctr, rMax, nx, &xLo, &xHi);
    int yLo, yHi; float_image_get_index_range(yctr, rMax, ny, &yLo, &yHi);
    
    double orad = rad + hwd;  /* Outer half-side. */
    double hwd2 = hwd*hwd;    /* Tip radius, squared. */

    auto float func_cross(double x, double y);

    float func_cross(double x, double y)
      { double dx = x - xctr;
        double dy = y - yctr;
        if (diagonal)
          { /* Rotate by 45 degrees: */
            double tx = M_SQRT1_2 * (dx + dy);
            double ty = M_SQRT1_2 * (dy - dx); 
            dx = tx; dy = ty;
          }
        dx = fabs(dx);
        dy = fabs(dy);
        if (dx < dy) { double t = dx; dx = dy; dy = t; }
        if ((dx < rad) && (dy < hwd))
          { return vdraw; }
        else if ((dx > orad) || (dy > hwd))
          { return NAN; }
        else 
          { /* Round tip: */
            dx -= rad;
            double dr2 = dx*dx + dy*dy;
            if (dr2 < hwd2) 
              { return vdraw; }
            else
              { return NAN; }
          }
      }

    float_image_paint_samples(A, c, xLo, xHi, yLo, yHi, func_cross, m);
  }
